#ifndef __COMPOSITEPATHALGEBRAS__ 
#define __COMPOSITEPATHALGEBRAS__ 

#include <iostream>
#include <sstream>
#include <string>
#include <limits>
#include "PathAlgebras.h"
#include "SetAlgebras.h"
#include "VectorAlgebras.h"

using namespace std;

template<class Type1, class Type2>
class CompositePathAlgebra {
//CompositePathAlgebra basic type
	Type1 element1;
	Type2 element2;
public:
	CompositePathAlgebra(const Type1 &e1 = Type1().Zero(), const Type2 &e2 = Type2().Zero()) : element1(e1), element2(e2) {
	}
	
	const CompositePathAlgebra& operator+ (const CompositePathAlgebra &s) const {
		static CompositePathAlgebra *r = new CompositePathAlgebra();
		if (element1 + s.element1 == element1)    return (*this);
		if (element1 + s.element1 == s.element1)  return s;
			r->element1 = element1 + s.element1;
			r->element2 = element2 + s.element2;
			return *r;
	}
	const CompositePathAlgebra& operator* (const CompositePathAlgebra &s) const {
		static CompositePathAlgebra *r = new CompositePathAlgebra();
		r->element1 = element1 * s.element1;
		r->element2 = element2 * s.element2;
		return *r;
	}
	bool operator==(const CompositePathAlgebra &s) const {
		if (element1 == s.element1 && element2 == s.element2) return true;
		return false;
	}
	virtual bool operator<=(const CompositePathAlgebra &s)	const {
		return   ( element1 <= s.element1 ); 
	}
	string ToString() const {
		return "|" + element1.ToString() + "*" + element2.ToString() + "|";
	}
	string ToDelimitedString(char delim = ' ') const {
		return element1.ToDelimitedString() + delim + element2.ToDelimitedString();
	}
	virtual void FromString(string & s) {
		element1 = element1.Zero();
		element2 = element2.Zero();
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		char c;
		is >> c;
		if (c != '|') exit(1);
		s = s.substr(1); is.str(s);
		element1.FromString(s); is.str(s);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		is >> c;
		if (c != '*') exit(1);
		s = s.substr(1); is.str(s);
		element2.FromString(s); is.str(s);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		is >> c;
		if (c != '|') exit(1);
		s = s.substr(1); is.str(s);
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		element1 = element1.Zero();
		element2 = element2.Zero();
		element1.FromDelimitedString(s);
		element2.FromDelimitedString(s);
	}

	const pair<Type1,Type2>& GetElement() const {
		static pair<Type1,Type2> t;
		t.first = element1;
		t.second = element2;
		return t;
	}
	CompositePathAlgebra Zero() const {
		return CompositePathAlgebra();
	}
	CompositePathAlgebra One() const {
		return CompositePathAlgebra(Type1().One(), Type2().One());
	}
};

template<class Type1, class Type2>
ostream& operator<< (ostream &os, const CompositePathAlgebra<Type1,Type2> &s) {
	os << s.ToString();
	return os;
}

template<class Type1, class Type2>
PathAlgebraBaseType<CompositePathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > > 
	CompositeAlgebra(const PathAlgebraBaseType<Type1> &e1, const PathAlgebraBaseType<Type2> &e2) {
	return Algebra(CompositePathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > (e1,e2) );
}

template<class Type1, class Type2>
PathAlgebraBaseType<CompositePathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > > 
	C(const PathAlgebraBaseType<Type1> &e1, const PathAlgebraBaseType<Type2> &e2) {
	return Algebra(CompositePathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > (e1,e2) );
}

template<class Type1, class Type2, int n>
Type1 Max (const PathAlgebraBaseType<CompositePathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1,n> >,PathAlgebraBaseType<Type2> > >  &e) {
	return Max(e.GetElement().GetElement().first);
}

template<class Type1, class Type2, int n>
Type1 Min (const PathAlgebraBaseType<CompositePathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1,n> >,PathAlgebraBaseType<Type2> > >  &e) {
	return Min(e.GetElement().GetElement().first);
}

template<class Type1, class Type2, int n>
auto MaxMinAlgebra(const PathAlgebraBaseType<SetPathAlgebra<PathAlgebraBaseType<CompositePathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1,n> >,PathAlgebraBaseType<Type2> > > > >  &e) {
 	auto sa = e.GetElement(); //SetAlgebra
  	auto lav = sa.GetElement(); // list of PathAlgebras
	if (lav.begin() == lav.end() ) {
		cerr << "MaxMin error: list empty." << endl;
		exit(1);
	}
	auto i = lav.begin(); // iterator to PathAlgebra
	auto maxi = i;
	Type1 maxmin = Min(*i);
	++i;
	while (i != lav.end()) {
		auto nextmax = Min(*i);
		if (!(nextmax <= maxmin)) {
			maxmin = nextmax;
			maxi = i;
		}
		++i;
	}
	return *maxi;
}

template<class Type1, class Type2, int n>
auto MaxMinValue(const PathAlgebraBaseType<SetPathAlgebra<PathAlgebraBaseType<CompositePathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1,n> >,PathAlgebraBaseType<Type2> > > > >  &e) {
	return Min( MaxMinAlgebra(e) );
}


#endif

