#ifndef __SETPATHALGEBRAS__ 
#define __SETPATHALGEBRAS__ 

#include <iostream>
#include <sstream>
#include <string>
#include <limits>
#include <set>
#include <list>
#include "PathAlgebras.h"
#include "VectorAlgebras.h"

using namespace std;


template<class Type>
class SetPathAlgebra {
//SetPathAlgebra basic type
//	set<Type> element;
	list<Type> element;
public:
	const static SetPathAlgebra zero;
	const static SetPathAlgebra one;
	SetPathAlgebra() {
	}
	
	SetPathAlgebra(const Type &e) {
		Insert(e);
	}

	const SetPathAlgebra& operator+ (const SetPathAlgebra &s) const {
		static SetPathAlgebra *r = new SetPathAlgebra();
		r->element.clear();
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			r->Insert(*i);
		}
		for (auto i = s.element.begin(); i !=s.element.end(); ++i ) {
			r->Insert(*i);
		}
		return *r;
	}
	const SetPathAlgebra& operator* (const SetPathAlgebra &s) const {
		static SetPathAlgebra *r = new SetPathAlgebra();
		r->element.clear();
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			for (auto j = s.element.begin(); j !=s.element.end(); ++j ) {
				r->Insert((*i) * (*j));
			}
		}
		return *r;
	}
	bool operator==(const SetPathAlgebra &s) const {
		if (element.size() != s.element.size()) return false;
		if (element.size() == 0 || s.element.size() == 0) return false;
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			bool r = false;
			for (auto j = s.element.begin(); j !=s.element.end(); ++j ) {
				if (*i == *j) {
					r = true;
					break;
				}
			}
			if (r == false) return false;
		}
		return true;
		return false;
	}

	bool lex_LE(const SetPathAlgebra &s) const {
		auto i = element.begin();
		auto j = s.element.begin();
		while (i != element.end() && j != s.element.end()) {
			if (i->first.lex_LE(j->first)) return true;
			else if (j->first.lex_LE(i->first)) return false;
			++i; ++j;
		}
		return false; 
	}


	string ToString() const {
		string s = "{";
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			s += (*i).ToString();
			s += ',';
		}
		if (s[s.size()-1] == ',') s[s.size()-1] = '}';
		else s += "}";
		return s;
	}
	string ToDelimitedString(char delim = ' ') const {
		string s = "";
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			s += i->ToDelimitedString() + "\t";
		}
		return s;
	}
	virtual void FromString(string & s) {
		element.clear();
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		char c;
		is >> c;
		if (c != '{') exit(1);
		s = s.substr(1); is.str(s);
		do {
			Type t;
			t.FromString(s); is.str(s);
			Insert(t);
			it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
	  		s.erase( s.begin() , it);
			is >> c;
			s = s.substr(1); is.str(s);
			if (c == ',') {
				continue;
			}
			else break;
		} while (true);
		if (c == '}') return;
		exit(1);		
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		element.clear();
		Type t;
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
		s.erase( s.begin() , it);
		t.FromDelimitedString(s); 
		Insert(t);
	}

	const list<Type>& GetElement() const {
		return element;
	}
	
	list<Type>& Get() {
		return element;
	}
	
	void SetElement(const set<Type>& d)  {
		element = d;
	}

	void Insert(const Type &e)  {
		if (dominated(e)) { //if e is strictly dominated, don't insert it
			return;
		}
		eff(e); //erase all elements that e strictly dominates
		element.push_front(e);
	}

	void Erase(const Type &e)  {
		element.erase(e);
	}

	SetPathAlgebra Zero() const {
		return zero;
	}
	
	SetPathAlgebra One() const {
		return one;
	}
	
	void eff() {
		for (auto i = element.begin(); i !=element.end(); ++i) {
			for (auto j = element.begin(); j !=element.end(); ) {
				if (i == j) {
					++j;
					continue;
				}
//				if (*j <= *i) {
				if (*j < *i) {
					j = element.erase(j);
				}
				else ++j;
			}
		}
	}
	
	bool dominated(const Type &e) {
		for (auto j = element.begin(); j !=element.end(); ++j) {
			if (e < *j) return true;
		}
		return false;
	}
	
	void eff(const Type &e) {
		for (auto j = element.begin(); j !=element.end(); ) {
			if (*j <= e) {
				j = element.erase(j);
			}
			else ++j;
		}
	}
	
};

template<class Type>
const SetPathAlgebra<Type> SetPathAlgebra<Type>::zero ;

template<class Type>
const SetPathAlgebra<Type> SetPathAlgebra<Type>::one (Type::one) ;

template<class Type>
ostream& operator<< (ostream &os, const SetPathAlgebra<Type> &s) {
	os << s.ToString();
	return os;
}


template<class Type>
PathAlgebraBaseType<SetPathAlgebra<PathAlgebraBaseType<Type> > > SetAlgebra(const PathAlgebraBaseType<Type> &e) {
	return Algebra(SetPathAlgebra<PathAlgebraBaseType<Type> > () );
}

template<class Type>
PathAlgebraBaseType<SetPathAlgebra<PathAlgebraBaseType<Type> > > S(const PathAlgebraBaseType<Type> &e) {
	return Algebra(SetPathAlgebra<PathAlgebraBaseType<Type> > () );
}

template<class Type>
auto MaxMinAlgebra(const PathAlgebraBaseType<SetPathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type> > > > &e) {
 	auto sa = e.GetElement(); //SetAlgebra
  	auto lav = sa.GetElement(); // list of PathAlgebras
	if (lav.begin() == lav.end() ) {
		cerr << "MaxMin error: list empty." << endl;
		exit(1);
	}
	auto i = lav.begin(); // iterator to PathAlgebra
	auto maxi = i;
	Type maxmin = Min(*i);
	++i;
	while (i != lav.end()) {
		auto nextmax = Min(*i);
		if (!(nextmax <= maxmin)) {
			maxmin = nextmax;
			maxi = i;
		}
		++i;
	}
	return *maxi;
}

template<class Type>
auto MaxMinValue(const PathAlgebraBaseType<SetPathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type> > > > &e) {
	return Min( MaxMinAlgebra(e) );
}


#endif
