#ifndef __VECTORPATHALGEBRAS__ 
#define __VECTORPATHALGEBRAS__ 

#include <iostream>
#include <sstream>
#include <string>
#include <limits>
#include "Timer.h"

using namespace std;

template<class Type1, class Type2>
class VectorPathAlgebra {
//VectorPathAlgebra basic type
	Type1 element1;
	Type2 element2;
public:
	static const VectorPathAlgebra zero;
	static const VectorPathAlgebra one;

	VectorPathAlgebra(const Type1 &e1 = Type1::zero, const Type2 &e2 = Type2::zero) : element1(e1), element2(e2) {
	}
	
	const VectorPathAlgebra& operator+ (const VectorPathAlgebra &s) const {
		static VectorPathAlgebra *r = new VectorPathAlgebra();
		r->element1 = element1 + s.element1;
		r->element2 = element2 + s.element2;
		return *r;
	}
	const VectorPathAlgebra& operator* (const VectorPathAlgebra &s) const {
		static VectorPathAlgebra *r = new VectorPathAlgebra();
		r->element1 = element1 * s.element1;
		r->element2 = element2 * s.element2;
		return *r;
	}
	bool operator==(const VectorPathAlgebra &s) const {
		if (element1 == s.element1 && element2 == s.element2) return true;
		return false;
	}

	bool operator<=(const VectorPathAlgebra &s) const {
		if (element1 <= s.element1 && element2 <= s.element2) return true;
		return false;
	}

	string ToString() const {
		return "(" + element1.ToString() + "," + element2.ToString() + ")";
	}
	string ToDelimitedString(char delim = ' ') const {
		return element1.ToDelimitedString() + delim + element2.ToDelimitedString();
	}
	void FromString(string & s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		char c;
		is >> c;
		if (c != '(') exit(1);
		s = s.substr(1); is.str(s);
		element1.FromString(s); is.str(s);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		is >> c;
		if (c != ',') exit(1);
		s = s.substr(1); is.str(s);
		element2.FromString(s); is.str(s);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		is >> c;
		if (c != ')') exit(1);
		s = s.substr(1); is.str(s);
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		element1.FromDelimitedString(s);
		element2.FromDelimitedString(s);
	}

	const VectorPathAlgebra& Zero() const {
		return zero;
	}
	const VectorPathAlgebra& One() const {
		return one;
	}
};

template<class Type1, class Type2>
const VectorPathAlgebra<Type1,Type2> VectorPathAlgebra<Type1,Type2>::zero(Type1::zero,Type2::zero) ;

template<class Type1, class Type2>
const VectorPathAlgebra<Type1,Type2> VectorPathAlgebra<Type1,Type2>::one (Type1::one,Type2::one) ;

template<class Type1, class Type2>
ostream& operator<< (ostream &os, const VectorPathAlgebra<Type1,Type2> &s) {
	os << s.ToString();
	return os;
}


template<class Type>
class VectorNPathAlgebra {
//VectorNPathAlgebra basic type
	vector<Type> element;
public:
	static const VectorNPathAlgebra zero;
	static const VectorNPathAlgebra one;
	VectorNPathAlgebra(const Type& e = Type::zero) {
		element.push_back(e);
	}
	VectorNPathAlgebra(const int n, const Type& e = Type::zero) {
		for (int i = 0; i < n; i++) element.push_back(e);
	}
	const VectorNPathAlgebra& operator+ (const VectorNPathAlgebra &s) const {
		if (element.size() == s.element.size()) {
			static VectorNPathAlgebra *r = new VectorNPathAlgebra();
			r->clear();
			for (int i=0; i<element.size(); ++i) {
				r->element.push_back(element[i] + s.element[i]);
			}
			return *r;
		}
		PrintErrorAndExit("VectorNPathAlgebra& operator+ Error: Vectors should be of equal lenght"); 
		if (IsZero()) return s;
		else if (s.IsZero()) return (*this);
		else if (IsOne()) return (*this);
		else if (s.IsOne()) return s; 
		PrintErrorAndExit("VectorNPathAlgebra& operator+ Error"); 
	}
	const VectorNPathAlgebra& operator* (const VectorNPathAlgebra &s) const {
		static VectorNPathAlgebra *r = new VectorNPathAlgebra();
		if (element.size() == s.element.size()) {
			r->element.clear();
			for (int i=0; i<element.size(); ++i) {
				r->element.push_back(element[i] * s.element[i]);
			}
			return *r;
		}
		PrintErrorAndExit("VectorNPathAlgebra& operator+ Error: Vectors should be of equal lenght"); 
		if (IsZero()) return (*this); 
		else if (s.IsZero()) return s; 
		else if (IsOne()) return s;
		else if (s.IsOne()) return (*this); 
		PrintErrorAndExit("VectorNPathAlgebra& operator* error:");
		return *r;
	}
	bool operator==(const VectorNPathAlgebra &s) const {
		if ( element.size() != s.element.size() ) return false;
		for (int i=0; i<element.size(); ++i) if (! (element[i] == s.element[i]) ) return false;
		return true;
	}

	bool operator<=(const VectorNPathAlgebra &s) const {
		if (element.size() != s.element.size()) PrintErrorAndExit("VectorNPathAlgebra::operator<= Error");
		for (int i=0; i<element.size(); ++i) {
			if (! (element[i] <= s.element[i]) ) return false;
		}
		return true;
	}

	bool lex_LE(const VectorNPathAlgebra &s) const {
		if (element.size() != s.element.size()) PrintErrorAndExit("VectorNPathAlgebra::lex_LE Error");
		for (int i=0; i<element.size(); ++i) {
			if (element[i].lex_LE(s.element[i]) ) return true;
			else if (s.element[i].lex_LE(element[i]) ) return false;
		}
		return false; 
	}

	bool operator>=(const VectorNPathAlgebra &s) const {
		if (element.size() != s.element.size()) PrintErrorAndExit("VectorNPathAlgebra::operator>= Error");
		for (int i=0; i<element.size(); ++i) if (element[i] < s.element[i]) return false;
		return true;
	}

	bool IsZero() const { 
		if (element.size() == 0) PrintErrorAndExit("VectorNPathAlgebra::IsZero Error: element.size() == 0");
		for (int i = 0; i < element.size(); ++i) {
			if ( !(element[i] == Type::zero) ) return false;
		}
		return true;
	}

	bool IsOne() const { 
		if (element.size() == 0) PrintErrorAndExit("VectorNPathAlgebra::IsOne Error: element.size() == 0");
		for (int i = 0; i < element.size(); ++i) {
			if ( !(element[i] == Type::one) ) return false;
		}
		return true;
	}

	string ToString() const {
		string s = "(";
		for (int i=0; i<element.size(); ++i) s += element[i].ToString() + ",";
		if (s[s.size()-1] == ',') s[s.size()-1] = ')';
		else s += ")";
		return s;
	}
	string ToDelimitedString(char delim = ' ') const {
		string s;
		for (int i=0; i<element.size(); ++i) s += (element[i].ToDelimitedString() + delim);
		return s;
	}
	void FromString(string & s) {
		PrintErrorAndExit("VectorAlgebra: FromString: function not implemented.");
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		for (int i=0; i<element.size(); ++i) {
			element[i].FromDelimitedString(s);
		}
	}

	void FromDelimitedString(const char *pc) {
		string s (pc);
		FromDelimitedString(s);
	}

	const VectorNPathAlgebra& Zero() const {
		return zero;
	}

	const VectorNPathAlgebra& One() const {
		return one;
	}

	Type Max() const { 
		Type m = element[0];
		for (int i = 1; i < element.size(); i++) {
			if (!(element[i] <= m)) m = element[i];
		}
		return m;
	}

	Type Min() const { 
		Type m = element[0];
		for (int i = 1; i < element.size(); i++) {
			if (element[i] < m) m = element[i];
		}
		return m;
	}

};

template<class Type>
const VectorNPathAlgebra<Type> VectorNPathAlgebra<Type>::zero(Type::zero) ;

template<class Type>
const VectorNPathAlgebra<Type> VectorNPathAlgebra<Type>::one (Type::one) ;

template<class Type>
ostream& operator<< (ostream &os, const VectorNPathAlgebra<Type> &s) {
	os << s.ToString();
	return os;
}


#endif

